unit infinitedataset;

interface

uses
  DB, Classes, SysUtils, Windows, Forms, Contnrs, ComObj, ActiveX;

type
  TListDataSet = class (TDataSet)
  protected
    // record data and status
    FIsTableOpen: Boolean;
    FList: TObjectList;
    FRecordSize: Integer; // actual data + housekeeping
    FCurrent, FPos: Integer;
    // dataset virtual methods
    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function GetRecordSize: Word; override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalClose; override;
    procedure InternalDelete; override;
    procedure InternalFirst; override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalHandleException; override;
    procedure InternalLast; override;
    procedure InternalOpen; override;
//    procedure InternalPost; override;
//    procedure InternalInsert; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    function IsCursorOpen: Boolean; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function GetRecordCount: Integer; override;
    procedure SetRecNo(Value: Integer); override;
    function GetRecNo: Integer; override;
    // for specific subclasses
    procedure ReadListData; virtual; abstract;
  public
    constructor Create (Owner: TComponent); override;
    destructor Destroy; override;

  published
    // redeclared data set properties
    property Active;
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;
  end;

type
  PRecInfo = ^TRecInfo;
  TRecInfo = record
    Index: Integer;
    Bookmark: Longint;
    BookmarkFlag: TBookmarkFlag;
  end;

  TRandomDataset = class(TListDataSet)
  private
  protected
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalInitFieldDefs; override;
    procedure InternalInsert; override;
    procedure InternalPost; override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
    procedure ReadListData; override;
    function GetCanModify: Boolean; override;
  public
    function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
  end;

  TData = class
  public
    Name: string;
    OID: string;
    Value: Integer;
    constructor Create ();
  end;


implementation

function TListDataSet.AllocRecordBuffer: PChar;
begin
  Result := StrAlloc(fRecordSize);
end;

procedure TListDataSet.InternalInitRecord(Buffer: PChar);
begin
  FillChar(Buffer^, FRecordSize, 0);
end;

procedure TListDataSet.FreeRecordBuffer (var Buffer: PChar);
begin
  StrDispose(Buffer);
end;

procedure TListDataSet.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  PInteger(Data)^ := PRecInfo(Buffer).Bookmark;
end;

function TListDataSet.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
  Result := PRecInfo(Buffer).BookmarkFlag;
end;

function TListDataSet.GetRecNo: Integer;
begin
  Result := FPos;
end;

function TListDataSet.GetRecord(Buffer: PChar; GetMode: TGetMode;
  DoCheck: Boolean): TGetResult;
begin
  Result := grOK; // default
  case GetMode of
    gmNext: // move on
    begin
      if fCurrent < fList.Count - 1 then
        Inc (fCurrent)
      else
        // INFINITE SUPPORT
        // usual code:
        // Result := grEOF; // end of file
        // special code:
        fCurrent := 1;
      inc (fPos);
    end;
    gmPrior: // move back
    begin
      if fCurrent > 0 then
        Dec (fCurrent)
      else
        // Result := grBOF; // begin of file
        fCurrent := fList.Count - 1;
      Dec (fPos);
    end;
    gmCurrent: // check if empty
      if fCurrent >= fList.Count then
        Result := grEOF;
  end;

  if Result = grOK then // read the data
    with PRecInfo(Buffer)^ do
    begin
      Index := fCurrent;
      BookmarkFlag := bfCurrent;
      Bookmark := fCurrent;
    end;
end;

function TListDataSet.GetRecordSize: Word;
begin
  Result := 4; // actual data without house-keeping
end;

procedure TListDataSet.InternalAddRecord(Buffer: Pointer;
  Append: Boolean);
begin
  // todo: support adding items
end;

procedure TListDataSet.InternalClose;
begin
  // disconnet and destroy field objects
  BindFields (False);
  if DefaultFields then
    DestroyFields;
  // closed
  FIsTableOpen := False;
end;

procedure TListDataSet.InternalDelete;
begin
  // todo: support deleting
end;

procedure TListDataSet.InternalFirst;
begin
  FCurrent := -1; // crack!
  FPos := 1;
end;

procedure TListDataSet.InternalGotoBookmark(Bookmark: Pointer);
begin
  if (Bookmark <> nil) then
    FCurrent := Integer (Bookmark);
end;

procedure TListDataSet.InternalHandleException;
begin
  Application.HandleException(Self);
end;

{procedure TListDataSet.InternalInsert;
begin
  // todo: support inserting
end;}

procedure TListDataSet.InternalLast;
begin
  FCurrent := FList.Count; // crack!
end;

procedure TListDataSet.InternalOpen;
begin
  // initialize field definitions and create fields
  InternalInitFieldDefs;
  if DefaultFields then
    CreateFields;
  BindFields (True);

  // read directory data
  ReadListData;

  // initialize
  FRecordSize := sizeof (TRecInfo);
  FCurrent := -1;
  fPOs := 1;
  BookmarkSize := sizeOf (Integer);
  FIsTableOpen := True;
end;

{procedure TListDataSet.InternalPost;
begin
  // todo: subclass
end;}

procedure TListDataSet.InternalSetToRecord(Buffer: PChar);
begin
  FCurrent := PRecInfo(Buffer).Index;
end;

function TListDataSet.IsCursorOpen: Boolean;
begin
  Result := FIsTableOpen;
end;

procedure TListDataSet.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  PRecInfo(Buffer).Bookmark := PInteger(Data)^;
end;

procedure TListDataSet.SetBookmarkFlag(Buffer: PChar;
  Value: TBookmarkFlag);
begin
  PRecInfo(Buffer).BookmarkFlag := Value;
end;

procedure TListDataSet.SetRecNo(Value: Integer);
begin
  // don't know...
end;

constructor TListDataSet.Create(Owner: TComponent);
begin
  inherited;
  FList := TObjectList.Create (True); // owns objects
end;

destructor TListDataSet.Destroy;
begin
  inherited;
  FList.Free;
end;

{ TFileData }

const
  names: array [0..9] of string =
    ('John', 'Paul', 'Bill', 'Mark', 'Ray',
    'Dale', 'Frank', 'Denise', 'Ether', 'None');

constructor TData.Create;
var
  guid: TGuid;
begin
  Name := names [random (10)];
  CoCreateGUID (guid);
  OID := GUIDToString (guid);
  Value := Random (10000);
end;

{ TRandomDataset }

function TRandomDataset.GetCanModify: Boolean;
begin
  Result := False;
end;

function TRandomDataset.GetFieldData(Field: TField;
  Buffer: Pointer): Boolean;
var
  Data: TData;
  begin
  Data := fList [PRecInfo(ActiveBuffer).Index] as TData;
  case Field.Index of
    0: Move (fPos, Buffer^, sizeof (Integer));
    1: StrCopy (Buffer, pchar(Data.Name));
    2: StrCopy (Buffer, pchar(Data.OID));
    3: Move (Data.Value, Buffer^, sizeof (Integer));
  end; // case
  Result := True;
end;

procedure TRandomDataset.InternalAddRecord(Buffer: Pointer;
  Append: Boolean);
begin
  inherited;
    // TODO: support adding
end;

procedure TRandomDataset.InternalInitFieldDefs;
begin
  // field definitions
  FieldDefs.Clear;
  FieldDefs.Add ('ID', ftInteger);
  FieldDefs.Add ('Name', ftString, 30, True);
  FieldDefs.Add ('OID', ftString, 50, True);
  FieldDefs.Add ('Value', ftInteger);
end;

procedure TRandomDataset.InternalInsert;
begin
  inherited;
  // TODO: support editing
end;

procedure TRandomDataset.InternalPost;
begin
  inherited;
  // TODO: support editing
end;

procedure TRandomDataset.ReadListData;
var
  i: integer;
begin
  FList.Clear;
  for i := 1 to 10 do
    FList.Add (TData.Create);
end;

procedure TRandomDataset.SetFieldData(Field: TField; Buffer: Pointer);
begin
  inherited;
  // TODO: support editing
end;

function TListDataSet.GetRecordCount: Integer;
begin
  Result := high (Integer);
end;

initialization
  randomize;

end.

